from typing import Any

from ms_graph_client.exceptions import UnableToFindUserError
from ms_graph_client.graph_api_crud_base import GraphAPICRUDBASE
from ms_graph_client.graph_api_config import GraphAPIConfig
import requests


class Users(GraphAPICRUDBASE):
    def __init__(self, config: GraphAPIConfig):
        super().__init__(config=config)

    def get_user(self, upn: str) -> Any:
        "carpnick2@qkdw.onmicrosoft.com"
        try:
            res = self._get(url_part="/users/" + upn)
            return res
        except Exception as e:
            if e.__cause__.__class__ == requests.exceptions.HTTPError:
                if e.__cause__.response.status_code == 404:
                    raise UnableToFindUserError(upn) from e
            # Default reraise
            raise



from typing import Any

from ms_graph_client.services.users import Users as UsersCrud
import cachetools.func


class Users:
    def __init__(self, crud_client: UsersCrud):
        self._crud_client = crud_client

    @cachetools.func.ttl_cache(ttl=600)
    def get_user(self, upn: str) -> Any:
        res = self._crud_client.get_user(upn=upn)
        return res



from .applications import Applications as Applications
from .groups import Groups as Groups
from .users import Users as Users
from ms_graph_client.graph_api_config import GraphAPIConfig as GraphAPIConfig

class GraphAPI:
    groups: Groups
    applications: Applications
    users: Users
    def __init__(self, config: GraphAPIConfig) -> None: ...



class TooManyMatchingGroupsError(Exception):
    pass


class NoMatchingGroupError(Exception):
    pass


class UnableToFindApplicationError(Exception):
    pass


class UnableToFindUserError(Exception):
    pass



from typing import Any
import cachetools.func
from ms_graph_client.services.groups import Groups as GroupsCrud


class Groups:
    def __init__(self, crud_client: GroupsCrud):
        self._crud_client = crud_client

    @cachetools.func.ttl_cache(ttl=300)
    def is_group_assigned_to_app(self, app_service_principal_id: str, group_id: str) -> bool:
        return self._crud_client.is_group_assigned_to_app(
            app_service_principal_id=app_service_principal_id, group_id=group_id
        )

    @cachetools.func.ttl_cache(ttl=300)
    def get_by_name(self, group_name: str) -> Any:
        return self._crud_client.get_by_name(group_name=group_name)

    @cachetools.func.ttl_cache(ttl=300)
    def exists_by_name(self, group_name: str) -> bool:
        return self._crud_client.exists_by_name(group_name=group_name)

    @cachetools.func.ttl_cache(ttl=300)
    def list_group_members(self, group_id: str) -> list[Any]:
        return self._crud_client.list_group_members(group_id=group_id)

    @cachetools.func.ttl_cache(ttl=300)
    def is_member_of_group(self, group_id: str, object_id: str) -> bool:
        return self._crud_client.is_member_of_group(group_id=group_id, object_id=object_id)



from ms_graph_client.services.applications import Applications as ApplicationsCrud
from typing import Any

class Applications:
    _crud_client: ApplicationsCrud
    def __init__(self, crud_client: ApplicationsCrud) -> None: ...
    def get_by_application_name(self, app_name: str) -> Any: ...
    def get_service_principal_by_app_id(self, app_id: str) -> Any: ...
    def exists_by_obj_id(self, id: str) -> bool: ...
    def get_by_object_id(self, id: str) -> Any: ...



from typing import Any
from ms_graph_client.graph_api_crud_base import GraphAPICRUDBASE
from ms_graph_client.graph_api_config import GraphAPIConfig
from ms_graph_client.exceptions import UnableToFindApplicationError
from .groups import Groups
import datetime
import time


class Applications(GraphAPICRUDBASE):
    def __init__(self, config: GraphAPIConfig, group_obj: Groups):
        super().__init__(config=config)
        self._group_crud = group_obj

    def get_by_application_name(self, app_name: str) -> Any:
        app_find = self._get(url_part="/applications?" + "$filter=displayName eq '" + app_name + "'")
        total_length = len(app_find["value"])
        assert total_length == 1

        return app_find["value"][0]

    def get_service_principal_by_app_id(self, app_id: str) -> Any:
        sp_resp = self._get(url_part="/servicePrincipals(appId='" + app_id + "')")
        return sp_resp

    def exists_by_obj_id(self, id: str) -> bool:
        try:
            self.get_by_object_id(id)
            return True
        except UnableToFindApplicationError:
            return False

    def get_by_object_id(self, id: str) -> Any:
        # https://learn.microsoft.com/en-us/graph/api/application-get?view=graph-rest-1.0&tabs=http
        # Application.Read.All,
        # Application.ReadWrite.OwnedBy,
        # Application.ReadWrite.All

        res = self._get(url_part="/applications/" + id)
        return res

    def list_assignments_to_app(self, app_sp_id: str) -> list[Any]:
        assignments = []

        myresp = self._get(url_part="/servicePrincipals/" + app_sp_id + "/appRoleAssignedTo/")
        for val in myresp["value"]:
            assignments.append(val)

        js = myresp
        while "@odata.nextLink" in js:
            js = self._get(url_part=js["@odata.nextLink"])
            for temp_member in js["value"]:
                assignments.append(temp_member)

        return assignments

    def _stabilize_app_assignment(self, app_sp_id: str, group_id: str, should_be_assigned: bool) -> None:
        expected_total_times = 4
        total_times_found = 0
        sleep_time = 3

        print(datetime.datetime.now().isoformat() + " - Stabilization Started")
        while True:

            # Really inefficient way of doing it
            # Replacing with group data instead - group data will be far smaller than app searching.
            # for val in self.list_assignments_to_app(app_sp_id=app_sp_id):
            #     if val["principalId"] == principal_id:
            #         return True
            #
            # return False

            if self._group_crud.is_group_assigned_to_app(app_service_principal_id=app_sp_id, group_id=group_id):
                if should_be_assigned:
                    total_times_found += 1
                    if total_times_found >= expected_total_times:
                        break
                else:
                    total_times_found = 0
            else:
                if should_be_assigned:
                    total_times_found = 0
                else:
                    total_times_found += 1
                    if total_times_found >= expected_total_times:
                        break

            time.sleep(sleep_time)

        time.sleep(3)
        print(datetime.datetime.now().isoformat() + " - Stabilization Succeeded")

    def assign_group_to_app(self, app_name: str, group_id: str, with_stabilization: bool, appRole: str = "User") -> Any:

        # FIND THE APPLICATION we are going to assign to
        aws_app_json = self.get_by_application_name(app_name=app_name)

        # Get service principal of app
        aws_sp_json = self.get_service_principal_by_app_id(aws_app_json["appId"])
        aws_sp_id = aws_sp_json["id"]

        if self._group_crud.is_group_assigned_to_app(app_service_principal_id=aws_sp_id, group_id=group_id):
            raise Exception("GroupId:" + group_id + " is already assigned to application app name: " + app_name)

        # Get the right app Role to assign the group to - appRole Role
        aws_app_role_id = None
        for role in aws_sp_json["appRoles"]:
            if role["displayName"] == appRole:
                aws_app_role_id = role["id"]

        assert aws_app_role_id is not None

        # Do the assignment
        data = {"principalId": group_id, "resourceId": aws_sp_id, "appRoleId": aws_app_role_id}
        myresp = self._post(url_part="/servicePrincipals/" + aws_sp_id + "/appRoleAssignedTo", json=data)

        if with_stabilization:
            self._stabilize_app_assignment(app_sp_id=aws_sp_id, group_id=group_id, should_be_assigned=True)

        return myresp

    def unassign_group_to_app(
        self, app_name: str, assigned_to_id: str, group_id: str, with_stabilization: bool
    ) -> None:

        # FIND THE APPLICATION we are going to assign to
        aws_app_json = self.get_by_application_name(app_name=app_name)

        # Get service prinicpal of app
        aws_sp_json = self.get_service_principal_by_app_id(aws_app_json["appId"])
        aws_sp_id = aws_sp_json["id"]

        self._delete(url_part="/servicePrincipals/" + aws_sp_id + "/appRoleAssignedTo/" + assigned_to_id)

        if with_stabilization:
            self._stabilize_app_assignment(app_sp_id=aws_sp_id, group_id=group_id, should_be_assigned=False)



from ms_graph_client import GraphAPIConfig

class Generator:
    config: GraphAPIConfig

    def __init__(self, config: GraphAPIConfig): ...
    def user_url(self, user_obj_id: str) -> str: ...
    def service_principal_url(self, app_service_principal_id: str) -> str:...



from ms_graph_client.services.groups import Groups as GroupsCrud
from typing import Any

class Groups:
    _crud_client: GroupsCrud
    def __init__(self, crud_client: GroupsCrud) -> None: ...
    def is_group_assigned_to_app(self, app_service_principal_id: str, group_id: str) -> bool: ...
    def get_by_name(self, group_name: str) -> Any: ...
    def exists_by_name(self, group_name: str) -> bool: ...
    def list_group_members(self, group_id: str) -> list[Any]: ...
    def is_member_of_group(self, group_id: str, object_id: str) -> bool: ...



from .graph_api_config import GraphAPIConfig
import requests
from requests.adapters import HTTPAdapter, Retry
import cachetools.func
from typing import Optional, Any


class GraphAPICRUDBASE:
    def __init__(self, config: GraphAPIConfig, session: Optional[requests.Session] = None):
        self.config = config

        if session is None:
            self.session: requests.Session = self._default_session()
        else:
            self.session: requests.Session = session

    def _default_session(self) -> requests.Session:
        session = requests.Session()
        retry = Retry(
            connect=3,
            allowed_methods=["GET", "POST", "PATCH", "PUT", "DELETE"],
            status_forcelist=[429],
            backoff_factor=0.5,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    def _combine_url(self, url_part: str) -> str:
        if "https://" in url_part:
            return url_part
        else:
            return self.config.api_url + "/" + url_part.strip("/")

    # By default token expires in 3600 seconds - we are renewing 1/2 of the way through
    # Hard coding so we dont have to track time?
    def _non_cacheable_token_call(self) -> Any:
        """
        https://learn.microsoft.com/en-us/graph/auth-v2-service#4-get-an-access-token

        POST https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token HTTP/1.1
                Host: login.microsoftonline.com
                Content-Type: application/x-www-form-urlencoded

                client_id=535fb089-9ff3-47b6-9bfb-4f1264799865
                &scope=https%3A%2F%2Fgraph.microsoft.com%2F.default
                &client_secret=qWgdYAmab0YSkuL1qKv5bPX
                &grant_type=client_credentials

        """

        str = "https://login.microsoftonline.com/" + self.config.tenant_id + "/oauth2/v2.0/token"
        d = {
            "client_id": self.config.client_id,
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": self.config.client_secret,
            "grant_type": "client_credentials",
        }

        s = requests.post(url=str, data=d)
        s.raise_for_status()
        return s.json()

    @cachetools.func.ttl_cache(maxsize=5, ttl=1800)
    def _token_call(self) -> Any:
        return self._non_cacheable_token_call()

    def _response_has_json(self, response: requests.Response) -> bool:
        return response.headers.get("Content-Type", "").startswith("application/json")

    def validate_credentials(self) -> None:
        self._non_cacheable_token_call()

    def _call(
        self,
        url: str,
        body: Optional[Any],
        json: Optional[Any],
        method: str,
        extra_headers: Optional[Any] = None,
    ) -> Any:

        if extra_headers is None:
            extra_headers = {}

        # Pull in a fresh token
        t_resp = self._token_call()

        api_response = None

        if "ConsistencyLevel" not in extra_headers.keys():
            extra_headers.update({"ConsistencyLevel": "eventual"})

        try:
            api_response = self.session.request(
                url=url,
                method=method,
                headers={"Authorization": f'{t_resp["token_type"]} {t_resp["access_token"]}', **extra_headers},
                data=body,
                json=json,
            )

            api_response.raise_for_status()

            if self._response_has_json(api_response):
                return api_response.json()
            else:
                return api_response

        except requests.exceptions.RequestException as e:
            reason = f"Request failed. Reason [{e}]"
            if api_response is not None:
                if self._response_has_json(api_response):
                    reason += f"\n{api_response.json()}"
            raise Exception(reason) from e

    def _get(
        self, url_part: str, json: Optional[Any] = None, body: Optional[Any] = None, extra_headers: Optional[Any] = None
    ) -> Any:
        return self._call(
            url=self._combine_url(url_part=url_part), body=body, method="GET", extra_headers=extra_headers, json=json
        )

    def _post(
        self, url_part: str, json: Optional[Any] = None, body: Optional[Any] = None, extra_headers: Optional[Any] = None
    ) -> Any:
        return self._call(
            url=self._combine_url(url_part=url_part), body=body, method="POST", extra_headers=extra_headers, json=json
        )

    def _put(
        self, url_part: str, json: Optional[Any] = None, body: Optional[Any] = None, extra_headers: Optional[Any] = None
    ) -> Any:
        return self._call(
            url=self._combine_url(url_part=url_part), body=body, method="PUT", extra_headers=extra_headers, json=json
        )

    def _patch(
        self, url_part: str, json: Optional[Any] = None, body: Optional[Any] = None, extra_headers: Optional[Any] = None
    ) -> Any:
        return self._call(
            url=self._combine_url(url_part=url_part), body=body, method="PATCH", extra_headers=extra_headers, json=json
        )

    def _delete(
        self, url_part: str, json: Optional[Any] = None, body: Optional[Any] = None, extra_headers: Optional[Any] = None
    ) -> Any:
        return self._call(
            url=self._combine_url(url_part=url_part), body=body, method="DELETE", extra_headers=extra_headers, json=json
        )



from ms_graph_client.graph_api_config import GraphAPIConfig as GraphAPIConfig
from ms_graph_client.services.graph_api import GraphAPI as GraphAPI
from ms_graph_client.services_cache.graph_api import GraphAPI as GraphCacheAPI
from ms_graph_client.services.generator import Generator as Generator

__all__ = ["GraphAPIConfig", "GraphAPI", "GraphCacheAPI", "Generator"]



from ms_graph_client import GraphAPIConfig


class Generator:
    def __init__(self, config: GraphAPIConfig):
        self.config = config

    def user_url(self, user_obj_id: str) -> str:
        return self.config.api_url + "/users/" + user_obj_id

    def service_principal_url(self, app_service_principal_id: str) -> str:
        return self.config.api_url + "/servicePrincipals/" + app_service_principal_id
